/* $Id: freqspec.cpp 61 2006-07-31 21:15:25Z pjohnson $
  ------------------------------------------------------------------------
  This file is part of the PIIM package; please see:
    Johnson, PLF and Slatkin, M. 2006. "Inference of population
    genetic parameters in metagenomics: a clean look at messy data."
    Genome Research (in press).
 
  Copyright 2006 The Regents of the University of California.  All
  rights reserved.
 
  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.
 
  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 
  This software is distributed under the terms of Version 2 of the
  GNU General Public License as published by the Free Software
  Foundation.  If you have not received a copy of Version 2 of the
  GNU General Public License with this program, you may download the
  GPL from the following url: http://www.gnu.org/copyleft/gpl.html.

  ---------------------------------------------------------------------------
  Classes to generate various frequency spectra
  ---------------------------------------------------------------------------
*/

#include "freqspec.hpp"
#include "util.hpp"
#include <stdexcept>
#include <gsl/gsl_sf_expint.h>

/*---------------------------------------------------------------------------*/
// PRE : string containing whitespace-separated list of parameter values
// POST: values parsed & assigned to parameters
void CFreqSpectrum::SetParams(const string &paramStr) {
    istringstream iss(paramStr);
    
    double param;
    iss >> param;
    for (int i = 0;  iss;  ++i) {
        SetParam(i) = param;
        iss >> param;
    }
}

/*---------------------------------------------------------------------------*/
// PRE : # of minor alleles (m), depth (n)
// POST: Pr(m | frequency_spectrum)
double CFSNeutral::x_Calc (int m, int n) const {
    double f = HasParam(eFixed) ? GetParam(eFixed) : 0;
    if (m == 0) {
        double norm = 0;
        for (int i = 1;  i < n;  ++i) {
            norm += 1.0/i;
        }
        
        return f + 1 - GetParam(eTheta) * norm;
    }
    return (1-f) * GetParam(eTheta) / m;
}

/*---------------------------------------------------------------------------*/
// PRE : # of minor alleles (b), depth (n)
// POST: Pr(m | fs) as derived in Polanski & Kimmel (2003).
double CFSGrowth::x_W (int b, int j, int n) const {
    try {
        return m_WCache[b][j][n];
    } catch (exception) {} //missed cache

    double res;
    if (j == 2) {
        res = 6.0/(n + 1.);
    } else if (j == 3) {
        res = 30.0 * (n - 2.0*b) / ((n+1)*(n+2));
    } else {
        res = - ((double)(j-1)*(3 + 2*(j-2))*(n - j + 2) /
                 ((j-2)*(2*j-5)*(n+j-1))) * x_W(b,j-2,n) +
            ((double) (2*j-1)*(n-2*b) / ((j-2)*(n+j-1))) * x_W(b,j-1,n);
    }
    m_WCache.Set(b).Set(j).Set(n) = res;
    return res;
}

double CFSGrowth::x_Fac (int start, int stop) const {
    double r = 1;
    for (int i = start;  i >= stop;  --i) {
        r *= i;
    }
    return r;
}

double CFSGrowth::x_V (int j, int n) const {
    return (2*j - 1) * (1 + pow(-1,j)) * x_Fac(n,n-j+1)/x_Fac(n+j-1,n);
}

double CFSGrowth::x_E(int j) const {
    //rho=0 leads to overflow --> use very small instead
    double rho = GetParam(eGrowthRate) > 0 ? GetParam(eGrowthRate) : 1e-10;
    double x = (double) gsl_sf_choose(j,2)/rho;
    if (x < 50) {
        return -exp(x) * gsl_sf_expint_Ei(-x);
    }

    double r = 0;
    const unsigned int K = 10;
    for (unsigned int k = 1;  k < K;  ++k) {
        r += pow(-1,k) * gsl_sf_fact(k - 1) / pow(x,k);
    }
    r *= -1;
    
    return r;
}

/*---------------------------------------------------------------------------*/
// PRE : # of minor alleles (m), depth (n)
// POST: Pr(m | frequency_spectrum)
double CFSGrowth::x_Calc (int numMinor, int depth) const {
    double denom = 0;
    for (int j = 2;  j <= depth;  ++j) {
        denom += x_E(j) * x_V(j, depth);
    }

    //rho=0 leads to overflow --> use very small instead
    double rho = GetParam(eGrowthRate) > 0 ? GetParam(eGrowthRate) : 1e-10;
    //'denom' is also the expected total tree length (in units of r)
    //theta=4Nu -- want to scale down to 2Nu (can think of rho=2Nr)
    double segProb = GetParam(eTheta)/2 * denom/rho;
    if (numMinor == 0) {
        return 1.0 - segProb;
    }

    double num = 0;
    for (int j = 2;  j <= depth;  ++j) {
        num += x_E(j) * x_W(numMinor, j, depth);
    }


    return segProb * (num / denom);
}
